from pathlib import Path

from dagster_dg_core.utils import is_macos, is_windows


def get_claude_desktop_config_path() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "Claude" / "claude_desktop_config.json"
    elif is_macos():
        return (
            Path.home()
            / "Library"
            / "Application Support"
            / "Claude"
            / "claude_desktop_config.json"
        )
    else:
        return Path.home() / ".config" / "claude" / "claude_desktop_config.json"



import functools
import textwrap
from abc import abstractmethod
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    TypedDict,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

import click
from click.core import ParameterSource
from dagster_shared.match import match_type
from dagster_shared.merger import deep_merge_dicts
from dagster_shared.plus.config import load_config
from dagster_shared.record import record
from dagster_shared.seven import is_valid_module_pattern
from dagster_shared.utils import remove_none_recursively
from dagster_shared.utils.config import (
    DgConfigFileFormat,
    detect_dg_config_file_format,
    discover_config_file,
    does_dg_config_file_exist,
    get_dg_config_path,
)
from typing_extensions import Never, NotRequired, Required, Self, TypeAlias, TypeGuard

from dagster_dg_core.error import DgError, DgValidationError
from dagster_dg_core.utils import (
    exit_with_error,
    generate_tool_dg_cli_in_project_in_workspace_error_message,
    get_toml_node,
    has_toml_node,
    modify_toml,
)
from dagster_dg_core.utils.warnings import DgWarningIdentifier, emit_warning

if TYPE_CHECKING:
    import tomlkit
    import tomlkit.items

T = TypeVar("T")


def discover_workspace_root(path: Path) -> Optional[Path]:
    workspace_config_path = discover_config_file(
        path, lambda config: config["directory_type"] == "workspace"
    )
    return workspace_config_path.parent if workspace_config_path else None


@record
class DgConfigFileDiscoveryResult:
    root_path: Path
    workspace_root_path: Optional[Path] = None
    root_file_path: Optional[Path] = None
    root_validation_result: Optional["DgConfigValidationResult"] = None
    container_workspace_file_path: Optional[Path] = None
    container_workspace_validation_result: Optional["DgConfigValidationResult"] = None
    user_file_path: Optional[Path] = None
    user_config: Optional["DgRawCliConfig"] = None
    cli_config_warning: Optional[str]

    @property
    def has_root_file(self) -> bool:
        return self.root_file_path is not None

    @property
    def has_container_workspace_file(self) -> bool:
        return self.container_workspace_file_path is not None

    @property
    def has_user_file(self) -> bool:
        return self.user_file_path is not None

    @property
    def root_result(self) -> "DgConfigValidationResult":
        if not self.root_validation_result:
            raise DgError("No root file validation result available.")
        return self.root_validation_result

    @property
    def container_workspace_result(self) -> "DgConfigValidationResult":
        if not self.container_workspace_validation_result:
            raise DgError("No container workspace validation result available.")
        return self.container_workspace_validation_result

    @property
    def root_type(self) -> Optional[str]:
        if not self.root_file_path:
            return None
        return self.root_result.type

    @property
    def root_config(self) -> Optional["DgFileConfig"]:
        if not self.root_validation_result:
            return None
        return self.root_validation_result.config

    @property
    def container_workspace_config(self) -> Optional["DgWorkspaceFileConfig"]:
        if not self.container_workspace_validation_result:
            return None
        return cast("DgWorkspaceFileConfig", self.container_workspace_validation_result.config)


def discover_and_validate_config_files(path: Path) -> DgConfigFileDiscoveryResult:
    root_config_path = discover_config_file(path)
    workspace_config_path = discover_config_file(
        path, lambda x: bool(x.get("directory_type") == "workspace")
    )

    cli_config_warning: Optional[str] = None
    if root_config_path:
        root_path = root_config_path.parent
        root_file_validation_result = validate_dg_file_config(root_config_path)
        if workspace_config_path is None:
            workspace_root_path = None
            container_workspace_validation_result = None

        # Only load the workspace config if the workspace root is different from the first
        # detected root.
        elif workspace_config_path == root_config_path:
            workspace_root_path = workspace_config_path.parent
            container_workspace_validation_result = None
        else:
            workspace_root_path = workspace_config_path.parent
            container_workspace_validation_result = validate_dg_file_config(workspace_config_path)
            if (
                not root_file_validation_result.has_errors
                and "cli" in root_file_validation_result.config
            ):
                del root_file_validation_result.config["cli"]
                # We have to emit this _after_ we merge all configs to ensure we have the right
                # suppression list.
                cli_config_warning = generate_tool_dg_cli_in_project_in_workspace_error_message(
                    root_path, workspace_root_path
                )
    else:
        root_path = Path.cwd()
        workspace_root_path = None
        root_file_validation_result = None
        container_workspace_validation_result = None

    if has_dg_user_file_config():
        user_config = load_dg_user_file_config()
    else:
        user_config = None

    return DgConfigFileDiscoveryResult(
        root_path=root_path,
        workspace_root_path=workspace_root_path,
        root_file_path=root_config_path,
        root_validation_result=root_file_validation_result,
        container_workspace_file_path=workspace_config_path
        if root_path != workspace_root_path
        else None,
        container_workspace_validation_result=container_workspace_validation_result,
        user_file_path=get_dg_config_path() if has_dg_user_file_config() else None,
        user_config=user_config,
        cli_config_warning=cli_config_warning,
    )


# NOTE: The presence of dg.toml will cause pyproject.toml to be ignored for purposes of dg config.

_CLI_CONTEXT_CONFIG_KEY = "config"


def set_config_on_cli_context(cli_context: click.Context, config: "DgRawCliConfig") -> None:
    cli_context.ensure_object(dict)
    cli_context.obj[_CLI_CONTEXT_CONFIG_KEY] = config


def has_config_on_cli_context(cli_context: click.Context) -> bool:
    return _CLI_CONTEXT_CONFIG_KEY in cli_context.ensure_object(dict)


def get_config_from_cli_context(cli_context: click.Context) -> "DgRawCliConfig":
    cli_context.ensure_object(dict)
    return cli_context.obj[_CLI_CONTEXT_CONFIG_KEY]


# ########################
# ##### MAIN
# ########################


@dataclass
class DgConfig:
    cli: "DgCliConfig"
    project: Optional["DgProjectConfig"] = None
    workspace: Optional["DgWorkspaceConfig"] = None

    @classmethod
    def default(cls) -> Self:
        return cls(DgCliConfig.default())

    # Note that this function takes an argument called `container_workspace_file_config` instead of
    # just `workspace_file_config` because this is only to be passed when the root file config
    # points to a project inside of a workspace. If there is no package and the
    # root is itself a workspace, then `root_file_config` corresponds to the workspace and
    # `container_workspace_file_config` should be None.
    @classmethod
    def from_partial_configs(
        cls,
        root_file_config: Optional["DgFileConfig"] = None,
        container_workspace_file_config: Optional["DgWorkspaceFileConfig"] = None,
        command_line_config: Optional["DgRawCliConfig"] = None,
        user_config: Optional["DgRawCliConfig"] = None,
    ) -> Self:
        cli_partials: list[DgRawCliConfig] = []

        if container_workspace_file_config:
            if "cli" in container_workspace_file_config:
                cli_partials.append(container_workspace_file_config["cli"])
            workspace_config = DgWorkspaceConfig.from_raw(
                container_workspace_file_config["workspace"]
            )

        if root_file_config:
            if "cli" in root_file_config:
                cli_partials.append(root_file_config["cli"])

            if is_workspace_file_config(root_file_config):
                project_config = None
                workspace_config = DgWorkspaceConfig.from_raw(root_file_config["workspace"])
            elif is_project_file_config(root_file_config):
                project_config = DgProjectConfig.from_raw(root_file_config["project"])
                if not container_workspace_file_config:
                    workspace_config = None

        else:
            project_config = None
            workspace_config = None

        if command_line_config:
            cli_partials.append(command_line_config)

        all_cli_config = [user_config, *cli_partials] if user_config else cli_partials
        cli_config = (
            DgCliConfig.from_raw(*all_cli_config) if all_cli_config else DgCliConfig.default()
        )

        return cls(cli_config, project_config, workspace_config)  # pyright: ignore[reportPossiblyUnboundVariable]


# ########################
# ##### CLI
# ########################


@dataclass
class DgCliConfig:
    """CLI configuration for Dg.

    Args:
        verbose (bool): If True, log debug information.
        use_component_modules (list[str]): Specify a list of modules containing components.
            Any components retrieved from the remote environment will be filtered to only include those
            from these modules. This is useful primarily for testing, as it allows targeting of a stable
            set of test components.
    """

    verbose: bool = False
    use_component_modules: list[str] = field(default_factory=list)
    suppress_warnings: list["DgWarningIdentifier"] = field(default_factory=list)
    telemetry_enabled: bool = True

    @classmethod
    def default(cls) -> Self:
        return cls()

    @classmethod
    def from_raw(cls, *partials: "DgRawCliConfig") -> Self:
        merged = cast("DgRawCliConfig", functools.reduce(lambda acc, x: {**acc, **x}, partials))  # type: ignore
        return cls(
            verbose=merged.get("verbose", DgCliConfig.verbose),
            use_component_modules=merged.get(
                "use_component_modules",
                cls.__dataclass_fields__["use_component_modules"].default_factory(),
            ),
            suppress_warnings=merged.get(
                "suppress_warnings",
                cls.__dataclass_fields__["suppress_warnings"].default_factory(),
            ),
            telemetry_enabled=merged.get("telemetry", {}).get(
                "enabled", DgCliConfig.telemetry_enabled
            ),
        )


class RawDgTelemetryConfig(TypedDict, total=False):
    enabled: bool


# All fields are optional
class DgRawCliConfig(TypedDict, total=False):
    verbose: bool
    use_component_modules: Sequence[str]
    suppress_warnings: Sequence[DgWarningIdentifier]
    telemetry: RawDgTelemetryConfig


# ########################
# ##### PROJECT
# ########################


class DgRawBuildConfig(TypedDict):
    registry: Optional[str]
    directory: Optional[str]


def merge_build_configs(
    workspace_build_config: Optional[DgRawBuildConfig],
    project_build_config: Optional[DgRawBuildConfig],
) -> DgRawBuildConfig:
    project_dict = remove_none_recursively(project_build_config or {})
    workspace_dict = remove_none_recursively(workspace_build_config or {})
    return cast(
        "DgRawBuildConfig",
        deep_merge_dicts(workspace_dict, project_dict),
    )


def merge_container_context_configs(
    workspace_container_context_config: Optional[Mapping[str, Any]],
    project_container_context_config: Optional[Mapping[str, Any]],
) -> Mapping[str, Any]:
    # defer for import performance
    from dagster_cloud_cli.config import DagsterCloudConfigDefaultsMerger

    merger = DagsterCloudConfigDefaultsMerger()
    return merger.merge(
        {**workspace_container_context_config} if workspace_container_context_config else {},
        {**project_container_context_config} if project_container_context_config else {},
    )


@dataclass
class DgProjectConfig:
    root_module: str
    defs_module: Optional[str] = None
    code_location_target_module: Optional[str] = None
    code_location_name: Optional[str] = None
    registry_modules: list[str] = field(default_factory=list)

    @classmethod
    def from_raw(cls, raw: "DgRawProjectConfig") -> Self:
        return cls(
            root_module=raw["root_module"],
            defs_module=raw.get("defs_module", DgProjectConfig.defs_module),
            code_location_name=raw.get("code_location_name", DgProjectConfig.code_location_name),
            code_location_target_module=raw.get(
                "code_location_target_module",
                DgProjectConfig.code_location_target_module,
            ),
            registry_modules=raw.get(
                "registry_modules", cls.__dataclass_fields__["registry_modules"].default_factory()
            ),
        )


class DgRawProjectConfig(TypedDict):
    root_module: Required[str]
    defs_module: NotRequired[str]
    code_location_target_module: NotRequired[str]
    code_location_name: NotRequired[str]
    registry_modules: NotRequired[list[str]]


# ########################
# ##### WORKSPACE
# ########################


@dataclass
class DgWorkspaceConfig:
    projects: list["DgWorkspaceProjectSpec"]
    scaffold_project_options: "DgWorkspaceScaffoldProjectOptions"

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceConfig") -> Self:
        projects = [DgWorkspaceProjectSpec.from_raw(spec) for spec in raw.get("projects", [])]
        scaffold_project_options = DgWorkspaceScaffoldProjectOptions.from_raw(
            raw.get("scaffold_project_options", {})
        )
        return cls(projects, scaffold_project_options)


class DgRawWorkspaceConfig(TypedDict, total=False):
    projects: list["DgRawWorkspaceProjectSpec"]
    scaffold_project_options: "DgRawWorkspaceNewProjectOptions"


@dataclass
class DgWorkspaceProjectSpec:
    path: Path

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceProjectSpec") -> Self:
        return cls(
            path=Path(raw["path"]),
        )


class DgRawWorkspaceProjectSpec(TypedDict, total=False):
    path: Required[str]


@dataclass
class DgWorkspaceScaffoldProjectOptions:
    use_editable_dagster: Union[str, bool] = False

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceNewProjectOptions") -> Self:
        return cls(
            use_editable_dagster=raw.get(
                "use_editable_dagster", DgWorkspaceScaffoldProjectOptions.use_editable_dagster
            ),
        )

    # This is here instead of on `DgRawWorkspaceNewProjectOptions` because TypedDict can't have
    # methods.
    @classmethod
    def get_raw_from_cli(
        cls,
        use_editable_dagster: Optional[str],
    ) -> "DgRawWorkspaceNewProjectOptions":
        raw_scaffold_project_options: DgRawWorkspaceNewProjectOptions = {}
        if use_editable_dagster:
            raw_scaffold_project_options["use_editable_dagster"] = (
                True if use_editable_dagster == "TRUE" else use_editable_dagster
            )
        return raw_scaffold_project_options


class DgRawWorkspaceNewProjectOptions(TypedDict, total=False):
    use_editable_dagster: Union[str, bool]


# ########################
# ##### CLI CONFIG
# ########################


def normalize_cli_config(
    cli_options: Mapping[str, object], cli_context: click.Context
) -> DgRawCliConfig:
    # Remove any options that weren't explicitly provided.
    filtered_options = {
        key: value
        for key, value in cli_options.items()
        if cli_context.get_parameter_source(key) != ParameterSource.DEFAULT
    }
    return _validate_cli_config(filtered_options)


def _validate_cli_config(cli_opts: Mapping[str, object]) -> DgRawCliConfig:
    try:
        for key, type_ in DgRawCliConfig.__annotations__.items():
            _validate_cli_config_setting(cli_opts, key, type_)
        _validate_cli_config_no_extraneous_keys(cli_opts)
    except DgValidationError as e:
        _raise_cli_config_validation_error(str(e))
    return cast("DgRawCliConfig", cli_opts)


def _validate_cli_config_setting(cli_opts: Mapping[str, object], key: str, type_: type) -> None:
    if key in cli_opts and not match_type(cli_opts[key], type_):
        raise DgValidationError(f"`{key}` must be a {type_.__name__}.")


def _validate_cli_config_no_extraneous_keys(cli_opts: Mapping[str, object]) -> None:
    extraneous_keys = [k for k in cli_opts.keys() if k not in DgRawCliConfig.__annotations__]
    if extraneous_keys:
        raise DgValidationError(f"Unrecognized fields:\n    {extraneous_keys}")


def _raise_cli_config_validation_error(message: str) -> None:
    raise DgError(f"Error in CLI options:\n    {message}")


# ########################
# ##### FILE CONFIG
# ########################

# The Dg*FileConfig classes wrap config extracted from a config file. This may be either a dg.toml
# file or a pyproject.toml file. For `dg.toml`, the config is defined a the top level. For
# pyproject.toml the config must be mounted on the `tool.dg` section. Either way, once the config
# has been parsed and extracted into a Dg*FileConfig class, it does not matter which file type it
# was sourced from.

DgFileConfigDirectoryType = Literal["workspace", "project"]


class DgWorkspaceFileConfig(TypedDict):
    directory_type: Required[Literal["workspace"]]
    workspace: Required[DgRawWorkspaceConfig]
    cli: NotRequired[DgRawCliConfig]


def is_workspace_file_config(config: "DgFileConfig") -> TypeGuard[DgWorkspaceFileConfig]:
    return config["directory_type"] == "workspace"


class DgProjectFileConfig(TypedDict):
    directory_type: Required[Literal["project"]]
    project: Required[DgRawProjectConfig]
    cli: NotRequired[DgRawCliConfig]


def is_project_file_config(config: "DgFileConfig") -> TypeGuard[DgProjectFileConfig]:
    return config["directory_type"] == "project"


DgFileConfig: TypeAlias = Union[DgWorkspaceFileConfig, DgProjectFileConfig]


@contextmanager
def modify_dg_toml_config(
    path: Path,
) -> Iterator[Union["tomlkit.TOMLDocument", "tomlkit.items.Table"]]:
    """Modify a TOML file as a tomlkit.TOMLDocument, preserving comments and formatting."""
    import tomlkit
    import tomlkit.items

    with modify_toml(path) as toml:
        if detect_dg_config_file_format(path) == "root":
            yield toml
        elif not has_toml_node(toml, ("tool", "dg")):
            raise KeyError(
                "TOML file does not have a tool.dg section. This is required for pyproject.toml files."
            )
        else:
            yield get_toml_node(toml, ("tool", "dg"), tomlkit.items.Table)


def load_dg_user_file_config(path: Optional[Path] = None) -> DgRawCliConfig:
    path = path or get_dg_config_path()
    contents = load_config(path).get("cli", {})

    return DgRawCliConfig(**{k: v for k, v in contents.items() if k != "plus"})


_OLD_USER_FILE_CONFIG_LOCATION = Path.home() / ".dg.toml"


def has_dg_user_file_config() -> bool:
    # Remove when we remove other deprecated stuff.
    if (Path.home() / ".dg.toml").exists():
        # We can't suppress this warning because we haven't loaded the config with the
        # suppress_warnings list yet.
        emit_warning(
            "deprecated_user_config_location",
            f"""
                Found config file ~/.dg.toml. This location for user config is no longer being read.
                Please move your configuration file to `{get_dg_config_path()}`.
            """,
            None,
            include_suppression_instruction=False,
        )
    return does_dg_config_file_exist()


def load_dg_root_file_config(
    path: Path, config_format: Optional[DgConfigFileFormat] = None
) -> DgFileConfig:
    return _load_dg_file_config(path, config_format)


def load_dg_workspace_file_config(path: Path) -> "DgWorkspaceFileConfig":
    config = _load_dg_file_config(path, None)
    if is_workspace_file_config(config):
        return config
    else:
        raise_file_config_validation_error("Expected a workspace configuration.", path)


def _load_dg_file_config(path: Path, config_format: Optional[DgConfigFileFormat]) -> DgFileConfig:
    validation_result = validate_dg_file_config(path, config_format)
    if validation_result.has_errors:
        raise_file_config_validation_error(validation_result.message, path)
    return validation_result.config


def validate_dg_file_config(
    path: Path, config_format: Optional[DgConfigFileFormat] = None
) -> "DgConfigValidationResult":
    """Validate a Dg config file at the given path."""
    import tomlkit
    import tomlkit.items

    toml = tomlkit.parse(path.read_text())
    config_format = config_format or detect_dg_config_file_format(path)
    if config_format == "root":
        raw_dict = toml.unwrap()
        path_prefix = None
    else:
        raw_dict = get_toml_node(toml, ("tool", "dg"), tomlkit.items.Table).unwrap()
        path_prefix = "tool.dg"
    return _DgConfigValidator(path_prefix).validate({k: v for k, v in raw_dict.items()})


_DgConfigErrorType: TypeAlias = Literal[
    "unrecognized_field",
    "missing_required_field",
    "invalid_value",
]


@record
class _DgConfigErrorRecord:
    """Record for errors encountered during Dg config validation."""

    @property
    @abstractmethod
    def message(self) -> str:
        """The error message to display."""


@record
class _DgConfigInvalidValueErrorRecord(_DgConfigErrorRecord):
    key: str
    expected_type_str: str
    value_str: str

    @property
    def message(self) -> str:
        return f"Invalid value for `{self.key}`:\n    Expected: {self.expected_type_str}\n    Received: {self.value_str}"


@record
class _DgConfigUnrecognizedFieldErrorRecord(_DgConfigErrorRecord):
    parent_key: str
    key: str

    @property
    def message(self) -> str:
        return f"Unrecognized field at `{self.parent_key}`:\n    {self.key}"


@record
class _DgConfigMissingRequiredFieldErrorRecord(_DgConfigErrorRecord):
    key: str
    expected_type_str: str

    @property
    def message(self) -> str:
        return f"Missing required value for `{self.key}`:\n    Expected: {self.expected_type_str}"


@record
class _DgConfigMiscellaneousErrorRecord(_DgConfigErrorRecord):
    description: str

    @property
    def message(self) -> str:
        return self.description


@record
class DgConfigValidationResult:
    raw_config: dict[str, Any]
    type: Optional[str]
    errors: list[_DgConfigErrorRecord] = field(default_factory=list)

    @property
    def message(self) -> str:
        """Get a message summarizing the validation result."""
        if not self.errors:
            return "Configuration is valid."
        else:
            return "\n".join(error.message for error in self.errors)

    @property
    def has_errors(self) -> bool:
        """Check if there are any validation errors."""
        return bool(self.errors)

    @property
    def config(self) -> Union["DgWorkspaceFileConfig", "DgProjectFileConfig"]:
        if self.errors:
            raise DgError(
                "Cannot access config when there are validation errors. "
                "Use the `errors` property to access the list of errors."
            )
        if self.type == "workspace":
            return cast("DgWorkspaceFileConfig", self.raw_config)
        elif self.type == "project":
            return cast("DgProjectFileConfig", self.raw_config)
        else:
            raise DgError("Unreachable: Invalid type in _DgConfigValidatorResult.")


class _DgConfigValidator:
    def __init__(self, path_prefix: Optional[str]) -> None:
        self.path_prefix = path_prefix
        self.errors: list[_DgConfigErrorRecord] = []

    def validate(self, raw_dict: dict[str, Any]) -> DgConfigValidationResult:
        self._normalize_deprecated_settings(raw_dict)
        self._validate_file_config_setting(
            raw_dict,
            "directory_type",
            Required[Literal["workspace", "project"]],
        )
        self._validate_dg_config_file_cli_section(raw_dict.get("cli", {}))

        directory_type = None
        if raw_dict.get("directory_type") == "workspace":
            self._validate_file_config_workspace_section(raw_dict.get("workspace", {}))
            self._validate_file_config_no_extraneous_keys(
                set(DgWorkspaceFileConfig.__annotations__.keys()), raw_dict, None
            )
            directory_type = raw_dict["directory_type"]
        elif raw_dict.get("directory_type") == "project":
            self._validate_file_config_project_section(raw_dict.get("project", {}))
            self._validate_file_config_no_extraneous_keys(
                set(DgProjectFileConfig.__annotations__.keys()), raw_dict, None
            )
            directory_type = raw_dict["directory_type"]
        return DgConfigValidationResult(
            raw_config=raw_dict,
            type=directory_type,
            errors=self.errors,
        )

    def _normalize_deprecated_settings(self, raw_dict: dict[str, Any]) -> None:
        """Normalize deprecated settings to the new format."""
        # We have to separately extract the warning suppression list since we haven't validated the
        # config yet.
        cli_section = raw_dict.get("cli", {})
        self._validate_file_config_setting(
            cli_section, "suppress_warnings", list[DgWarningIdentifier], "cli"
        )
        suppress_warnings = cast(
            "list[DgWarningIdentifier]", cli_section.get("suppress_warnings", [])
        )

        if has_toml_node(raw_dict, ("project", "python_environment")):
            full_key = self._get_full_key("project.python_environment")
            msg = textwrap.dedent(f"""
                Setting `{full_key}` is deprecated. This key can be removed.
            """).strip()
            emit_warning("deprecated_python_environment", msg, suppress_warnings)
            del raw_dict["project"]["python_environment"]

    def _validate_dg_config_file_cli_section(self, section: object) -> None:
        if not isinstance(section, dict):
            self._log_invalid_value_error("tool.dg.cli", get_type_str(dict), section)
            return
        for key, type_ in DgRawCliConfig.__annotations__.items():
            self._validate_file_config_setting(section, key, type_, "cli")
        self._validate_file_config_no_extraneous_keys(
            set(DgRawCliConfig.__annotations__.keys()), section, "cli"
        )

    def _validate_file_config_project_section(self, section: object) -> None:
        if not isinstance(section, dict):
            self._log_invalid_value_error("project", get_type_str(dict), section)
            return
        for key, type_ in DgRawProjectConfig.__annotations__.items():
            self._validate_file_config_setting(section, key, type_, "project")
        self._validate_file_config_no_extraneous_keys(
            set(DgRawProjectConfig.__annotations__.keys()), section, "project"
        )
        if "registry_modules" in section:
            for i, pattern in enumerate(section["registry_modules"]):
                if not is_valid_module_pattern(pattern):
                    full_key = self._get_full_key(f"project.registry_modules[{i}]")
                    self.errors.append(
                        _DgConfigInvalidValueErrorRecord(
                            key=full_key,
                            expected_type_str="A pattern consisting of '.'-separated segments that are either valid Python identifiers or wildcards ('*').",
                            value_str=str(pattern),
                        )
                    )

    def _validate_file_config_workspace_section(self, section: object) -> None:
        if not isinstance(section, dict):
            self._log_invalid_value_error("workspace", get_type_str(dict), section)
            return
        for key, type_ in DgRawWorkspaceConfig.__annotations__.items():
            if key == "projects":
                if self._validate_file_config_setting(section, key, list, "workspace"):
                    for i, spec in enumerate(section.get("projects") or []):
                        self._validate_file_config_workspace_project_spec(spec, i)
            elif key == "scaffold_project_options":
                self._validate_file_config_workspace_scaffold_project_options(
                    section.get("scaffold_project_options", {})
                )
            else:
                self._validate_file_config_setting(section, key, type_, "workspace")
        self._validate_file_config_no_extraneous_keys(
            set(DgRawWorkspaceConfig.__annotations__.keys()), section, "workspace"
        )

    def _validate_file_config_workspace_project_spec(self, section: object, index: int) -> None:
        if not isinstance(section, dict):
            self._log_invalid_value_error(
                f"workspace.projects[{index}]", get_type_str(dict), section
            )
            return
        for key, type_ in DgRawWorkspaceProjectSpec.__annotations__.items():
            self._validate_file_config_setting(section, key, type_, f"workspace.projects[{index}]")
        self._validate_file_config_no_extraneous_keys(
            set(DgRawWorkspaceProjectSpec.__annotations__.keys()),
            section,
            f"workspace.projects[{index}]",
        )

    def _validate_file_config_workspace_scaffold_project_options(self, section: object) -> None:
        if not isinstance(section, dict):
            self._log_invalid_value_error(
                "workspace.scaffold_project_options", get_type_str(dict), section
            )
            return
        for key, type_ in DgRawWorkspaceNewProjectOptions.__annotations__.items():
            self._validate_file_config_setting(
                section, key, type_, "workspace.scaffold_project_options"
            )
        self._validate_file_config_no_extraneous_keys(
            set(DgRawWorkspaceNewProjectOptions.__annotations__.keys()),
            section,
            "workspace.scaffold_project_options",
        )

    def _validate_file_config_no_extraneous_keys(
        self, valid_keys: set[str], section: Mapping[str, object], toml_path: Optional[str]
    ) -> None:
        extraneous_keys = [k for k in section.keys() if k not in valid_keys]
        parent_key = self._get_full_key(toml_path)
        for key in extraneous_keys:
            self.errors.append(
                _DgConfigUnrecognizedFieldErrorRecord(parent_key=parent_key, key=key)
            )

    # expected_type Any to handle typing constructs (`Literal` etc)
    def _validate_file_config_setting(
        self,
        section: Mapping[str, object],
        key: str,
        type_: Any,
        path_prefix: Optional[str] = None,
    ) -> bool:
        origin = get_origin(type_)
        is_required = origin is Required
        class_ = type_ if origin not in (Required, NotRequired) else get_args(type_)[0]

        error_type: Optional[_DgConfigErrorType] = None
        if is_required and key not in section:
            error_type = "missing_required_field"
        if key in section and not match_type(section[key], class_):
            error_type = "invalid_value"

        if error_type:
            full_key = f"{path_prefix}.{key}" if path_prefix else key
            type_str = get_type_str(class_)
            if error_type == "missing_required_field":
                self._log_missing_required_field_error(full_key, type_str)
            if error_type == "invalid_value":
                self._log_invalid_value_error(full_key, type_str, section[key])
            return False
        return True

    def _get_full_key(self, key: Optional[str]) -> str:
        if self.path_prefix:
            return f"{self.path_prefix}.{key}" if key else self.path_prefix
        return key if key else "<root>"

    def _log_missing_required_field_error(self, key: str, type_str: str) -> None:
        full_key = self._get_full_key(key)
        self.errors.append(
            _DgConfigMissingRequiredFieldErrorRecord(
                key=full_key,
                expected_type_str=type_str,
            )
        )

    def _log_invalid_value_error(self, key: str, type_str: str, value: object) -> None:
        full_key = self._get_full_key(key)
        self.errors.append(
            _DgConfigInvalidValueErrorRecord(
                key=full_key,
                expected_type_str=type_str,
                value_str=str(value),
            )
        )


def raise_file_config_validation_error(message: str, file_path: Path) -> Never:
    exit_with_error(
        textwrap.dedent(f"""
        Errors found in configuration file at:
            {file_path}
        """)
        + "\n"
        + message,
        do_format=False,
    )


# expected_type Any to handle typing constructs (`Literal` etc)
def get_type_str(t: Any) -> str:
    origin = get_origin(t)
    if origin is None:
        # It's a builtin or normal class
        if hasattr(t, "__name__"):
            return t.__name__  # e.g. 'int', 'bool'
        return str(t)  # Fallback
    else:
        # It's a parametric type like list[str], Union[int, str], etc.
        args = get_args(t)
        arg_strs = sorted([get_type_str(a) for a in args])
        if origin is Union:
            return " | ".join(arg_strs)
        if origin is Literal:
            arg_strs = [f'"{a}"' for a in arg_strs]
            return " | ".join(arg_strs)
        else:
            return f"{_get_origin_name(origin)}[{', '.join(arg_strs)}]"


def _get_origin_name(origin: Any) -> str:
    if origin is Sequence:
        return "Sequence"  # avoid the collections.abc prefix
    else:
        return origin.__name__



import re
import sys
from collections.abc import Iterable, Mapping, Sequence, Set
from typing import TYPE_CHECKING, Any, Optional, Union

import dagster_shared.check as check
from dagster_shared.error import (
    SerializableErrorInfo,
    make_simple_frames_removed_hint,
    remove_system_frames_from_error,
)
from dagster_shared.serdes.objects import EnvRegistryKey, EnvRegistryObjectSnap
from dagster_shared.serdes.objects.package_entry import (
    EnvRegistryManifest,
    EnvRegistryObjectFeature,
)
from packaging.version import Version

from dagster_dg_core.utils.warnings import emit_warning

if TYPE_CHECKING:
    from dagster_dg_core.context import DgContext


class EnvRegistry:
    @staticmethod
    def from_dg_context(
        dg_context: "DgContext", extra_modules: Optional[Sequence[str]] = None
    ) -> "EnvRegistry":
        """Fetches the set of available registry objects. The default set includes everything
        discovered under the "dagster_dg_cli.registry_modules" entry point group in the target environment. If
        `extra_modules` is provided, these will also be searched for component types.
        """
        if dg_context.config.cli.use_component_modules:
            plugin_manifest = _load_module_registry_objects(
                dg_context, dg_context.config.cli.use_component_modules
            )
        else:
            plugin_manifest = _load_entry_point_registry_objects(dg_context)

        if extra_modules:
            deduped_extra_modules = set(extra_modules).difference(plugin_manifest.modules)
            plugin_manifest = plugin_manifest.merge(
                _load_module_registry_objects(dg_context, deduped_extra_modules)
            )

        # Only load project plugin modules if there is no entry point
        if dg_context.is_project and not dg_context.has_registry_module_entry_point:
            if dg_context.project_registry_modules:
                deduped_project_modules = set(dg_context.project_registry_modules).difference(
                    plugin_manifest.modules
                )
                plugin_manifest = plugin_manifest.merge(
                    _load_module_registry_objects(dg_context, deduped_project_modules)
                )

        if (
            dg_context.has_registry_module_entry_point
            and not dg_context.config.cli.use_component_modules
            and dg_context.default_registry_root_module_name not in plugin_manifest.modules
        ):
            emit_warning(
                "missing_dg_plugin_module_in_manifest",
                f"""
                Your package defines a `dagster_dg_cli.registry_modules` entry point, but this module was not
                found in the plugin manifest for the current environment. This means either that
                your project is not installed in the current environment, or that the entry point
                metadata was added after your module was installed. Python entry points are
                registered at package install time. Please reinstall your package into the current
                environment to ensure the entry point is registered.

                Entry point module: `{dg_context.default_registry_root_module_name}`
                """,
                suppress_warnings=dg_context.config.cli.suppress_warnings,
            )

        return EnvRegistry(plugin_manifest)

    def __init__(self, plugin_manifest: EnvRegistryManifest):
        self._plugin_manifest = plugin_manifest
        self._object_lookup = {key: obj for obj in plugin_manifest.objects for key in obj.all_keys}

    @staticmethod
    def empty() -> "EnvRegistry":
        return EnvRegistry(EnvRegistryManifest(modules=[], objects=[]))

    @property
    def modules(self) -> Sequence[str]:
        return self._plugin_manifest.modules

    def module_entries(self, module: str) -> Set[EnvRegistryKey]:
        return {obj.key for obj in self._plugin_manifest.objects if obj.key.package == module}

    # This differs from "modules" in that it is a list of root modules-- so if there are two entry
    # points foo.lib and foo.lib2, this will return ["foo"].
    @property
    def packages(self) -> Sequence[str]:
        return sorted({m.split(".")[0] for m in self._plugin_manifest.modules})

    def get_objects(
        self,
        package: Optional[str] = None,
        feature: Optional[EnvRegistryObjectFeature] = None,
    ) -> Sequence[EnvRegistryObjectSnap]:
        return [
            entry
            for entry in self._plugin_manifest.objects
            if (package is None or package == entry.key.package)
            and (feature is None or feature in entry.features)
        ]

    def get(self, key: EnvRegistryKey) -> EnvRegistryObjectSnap:
        """Resolves a library object within the scope of a given component directory."""
        return self._object_lookup[key]

    def has(self, key: EnvRegistryKey) -> bool:
        return key in self._object_lookup

    def keys(self) -> Iterable[EnvRegistryKey]:
        yield from sorted(
            (obj.key for obj in self._plugin_manifest.objects), key=lambda k: k.to_typename()
        )

    def items(self) -> Iterable[tuple[EnvRegistryKey, EnvRegistryObjectSnap]]:
        yield from ((obj.key, obj) for obj in self._plugin_manifest.objects)

    def __repr__(self) -> str:
        return f"<EnvRegistry {list(self.keys())}>"


def all_components_schema_from_dg_context(dg_context: "DgContext") -> Mapping[str, Any]:
    """Generate a schema for all components in the current environment."""
    from dagster.components.list import list_all_components_schema

    return list_all_components_schema(entry_points=True, extra_modules=())


# ########################
# ##### HELPERS
# ########################

MIN_DAGSTER_COMPONENTS_LIST_PLUGINS_VERSION = Version("1.10.12")


def _load_entry_point_registry_objects(
    dg_context: "DgContext",
) -> EnvRegistryManifest:
    return _fetch_plugin_manifest(entry_points=True, extra_modules=[])


def _load_module_registry_objects(
    dg_context: "DgContext", modules: Iterable[str]
) -> EnvRegistryManifest:
    modules_to_fetch = set(modules)
    objects: list[EnvRegistryObjectSnap] = []

    if modules_to_fetch:
        plugin_manifest = _fetch_plugin_manifest(
            entry_points=False, extra_modules=list(modules_to_fetch)
        )
        for module in modules_to_fetch:
            objects_for_module = [
                obj for obj in plugin_manifest.objects if obj.key.namespace == module
            ]
            objects.extend(objects_for_module)

    return EnvRegistryManifest(modules=list(modules), objects=objects)


# Prior to MIN_DAGSTER_COMPONENTS_LIST_PLUGINS_VERSION, the relevant command was named "list
# library" instead of "list plugins". It also output a list[EnvRegistryObjectSnap] instead
# of a EnvRegistryManifest. This function handles normalizing the output across versions. We
# can compute a EnvRegistryManifest from the list[EnvRegistryObjectSnap], but it
# won't include any modules that register an entry point but don't expose any plugin objects.
def _fetch_plugin_manifest(entry_points: bool, extra_modules: Sequence[str]) -> EnvRegistryManifest:
    from dagster.components.list import list_plugins
    from rich.console import Console
    from rich.panel import Panel

    result = list_plugins(entry_points=entry_points, extra_modules=extra_modules)

    if isinstance(result, SerializableErrorInfo):
        clean_result = remove_system_frames_from_error(
            result.cause,
            make_simple_frames_removed_hint(),
        )
        message_match = check.not_none(
            re.match(r"^\S+:\s+(Error loading entry point `(\S+?)`.*)", result.message)
        )
        header_message = message_match.group(1)
        entry_point_module = message_match.group(2)
        console = Console()
        console.print(header_message)
        console.line()
        panel = Panel(
            clean_result.to_string(),
            title=f"Entry point error ({entry_point_module})",
            expand=False,
        )
        console.print(panel)
        sys.exit(1)
    return result


def get_specified_env_var_deps(component_data: Mapping[str, Any]) -> set[str]:
    if not component_data.get("requirements") or "env" not in component_data["requirements"]:
        return set()
    return set(component_data["requirements"]["env"])


env_var_regex = re.compile(r"\{\{\s*env\(\s*['\"]([^'\"]+)['\"]\)\s*\}\}")
env_var_regex_dot_notation = re.compile(r"\{\{\s*env\.([^'\"]+)\s*\}\}")


def get_used_env_vars(data_structure: Union[Mapping[str, Any], Sequence[Any], Any]) -> set[str]:
    if isinstance(data_structure, Mapping):
        return set.union(set(), *(get_used_env_vars(value) for value in data_structure.values()))
    elif isinstance(data_structure, str):
        raw_result = set(env_var_regex.findall(data_structure)).union(
            set(env_var_regex_dot_notation.findall(data_structure))
        )
        return {var.strip() for var in raw_result}
    elif isinstance(data_structure, Sequence):
        return set.union(set(), *(get_used_env_vars(item) for item in data_structure))
    else:
        return set()



import textwrap
from collections.abc import Sequence
from typing import Literal, Optional

import click
from typing_extensions import TypeAlias

from dagster_dg_core.utils import format_multiline_str

DgWarningIdentifier: TypeAlias = Literal[
    "cli_config_in_workspace_project",
    "deprecated_user_config_location",
    "deprecated_python_environment",
    "deprecated_dagster_dg_library_entry_point",
    "missing_dg_plugin_module_in_manifest",
    "project_and_activated_venv_mismatch",
    # create-dagster
    "create_dagster_outdated",
]


def emit_warning(
    warning_id: DgWarningIdentifier,
    msg: str,
    suppress_warnings: Optional[Sequence[DgWarningIdentifier]],
    include_suppression_instruction: bool = True,
) -> None:
    if warning_id not in (suppress_warnings or []):
        formatted_main_msg = textwrap.dedent(msg).strip()
        suppression_instruction = format_multiline_str(f"""
            To suppress this warning, add "{warning_id}" to the `cli.suppress_warnings` list in
            your configuration.
        """)
        full_msg = (
            f"{formatted_main_msg}\n\n{suppression_instruction}\n"
            if include_suppression_instruction
            else formatted_main_msg
        )
        click.secho(
            full_msg,
            fg="yellow",
            err=True,
        )



from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple, Optional

import click
from dagster_shared.serdes.objects import EnvRegistryKey
from dagster_shared.yaml_utils import parse_yamls_with_source_position
from dagster_shared.yaml_utils.source_position import (
    LineCol,
    SourcePosition,
    SourcePositionTree,
    ValueAndSourcePositionTree,
)
from yaml.scanner import ScannerError

from dagster_dg_core.component import EnvRegistry, get_specified_env_var_deps, get_used_env_vars
from dagster_dg_core.context import DgContext
from dagster_dg_core.utils.check_utils import error_dict_to_formatted_error

if TYPE_CHECKING:  # defer for import performance
    from jsonschema import ValidationError

COMPONENT_FILE_SCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "attributes": {"type": "object"},
        "requirements": {
            "type": "object",
            "properties": {"env": {"type": "array", "items": {"type": "string"}}},
            "additionalProperties": False,
        },
        "template_vars_module": {"type": "string"},
        "post_processing": {
            "type": "object",
            "properties": {
                "assets": {"type": "array", "items": {"type": "object"}},
            },
            "additionalProperties": False,
        },
    },
    "additionalProperties": False,
}


def _scaffold_value_and_source_position_tree(
    filename: str, row: int, col: int
) -> ValueAndSourcePositionTree:
    return ValueAndSourcePositionTree(
        value=None,
        source_position_tree=SourcePositionTree(
            position=SourcePosition(
                filename=filename, start=LineCol(row, col), end=LineCol(row, col)
            ),
            children={},
        ),
    )


class ErrorInput(NamedTuple):
    object_key: Optional[EnvRegistryKey]
    error: "ValidationError"
    source_position_tree: ValueAndSourcePositionTree


def check_yaml(
    dg_context: DgContext,
    resolved_paths: Sequence[Path],
    validate_requirements: bool,
) -> bool:
    # defer for import performance
    from jsonschema import Draft202012Validator, ValidationError

    top_level_component_validator = Draft202012Validator(schema=COMPONENT_FILE_SCHEMA)

    validation_errors: list[ErrorInput] = []
    all_specified_env_var_deps = set()

    component_contents_by_key: dict[EnvRegistryKey, Any] = {}
    modules_to_fetch = set()
    for component_dir in dg_context.defs_path.rglob("*"):
        if resolved_paths and not any(
            path == component_dir or path in component_dir.parents for path in resolved_paths
        ):
            continue

        defs_yaml_path = component_dir / "defs.yaml"
        component_yaml_path = component_dir / "component.yaml"

        yaml_path = (
            defs_yaml_path
            if defs_yaml_path.exists()
            else component_yaml_path
            if component_yaml_path.exists()
            else None
        )

        if yaml_path:
            text = yaml_path.read_text()
            try:
                component_doc_trees = parse_yamls_with_source_position(
                    text, filename=str(yaml_path)
                )
            except ScannerError as se:
                validation_errors.append(
                    ErrorInput(
                        None,
                        ValidationError(f"Unable to parse YAML: {se.context}, {se.problem}"),
                        _scaffold_value_and_source_position_tree(
                            filename=str(yaml_path),
                            row=se.problem_mark.line + 1 if se.problem_mark else 1,
                            col=se.problem_mark.column + 1 if se.problem_mark else 1,
                        ),
                    )
                )
                continue

            # Validate each YAML document in multi-document files
            for component_doc_tree in component_doc_trees:
                if validate_requirements:
                    specified_env_var_deps = get_specified_env_var_deps(component_doc_tree.value)
                    used_env_vars = get_used_env_vars(component_doc_tree.value)
                    all_specified_env_var_deps.update(specified_env_var_deps)

                    if used_env_vars - specified_env_var_deps:
                        msg = (
                            "Component uses environment variables that are not specified in the component file: "
                            + ", ".join(sorted(used_env_vars - specified_env_var_deps))
                        )

                        validation_errors.append(
                            ErrorInput(
                                None,
                                ValidationError(
                                    msg,
                                    path=["requirements", "env"],
                                ),
                                component_doc_tree,
                            )
                        )

                # First, validate the top-level structure of the component file
                # (type and params keys) before we try to validate the params themselves.
                top_level_errs = list(
                    top_level_component_validator.iter_errors(component_doc_tree.value)
                )
                for err in top_level_errs:
                    validation_errors.append(ErrorInput(None, err, component_doc_tree))
                if top_level_errs:
                    continue

                raw_key = component_doc_tree.value.get("type")
                component_instance_module = dg_context.get_component_instance_module_name(
                    component_dir.name
                )
                qualified_key = (
                    f"{component_instance_module}{raw_key}" if raw_key.startswith(".") else raw_key
                )
                key = EnvRegistryKey.from_typename(qualified_key)
                component_contents_by_key[key] = component_doc_tree

                # Add every module referenced to be explicitly fetched. If we don't do this, only
                # modules that are explicitly declared as registry modules will work.
                # `from_dg_context()` on the registry ensures that modules aren't double-fetched.
                modules_to_fetch.add(key.namespace)

    # Fetch the local component types, if we need any local components
    component_registry = EnvRegistry.from_dg_context(
        dg_context, extra_modules=list(modules_to_fetch)
    )
    for key, component_doc_tree in component_contents_by_key.items():
        try:
            json_schema = component_registry.get(key).component_schema or {}

            v = Draft202012Validator(json_schema)
            for err in v.iter_errors(component_doc_tree.value.get("attributes", {})):
                validation_errors.append(ErrorInput(key, err, component_doc_tree))
        except KeyError:
            # No matching component type found
            validation_errors.append(
                ErrorInput(
                    None,
                    ValidationError(f"Component type '{key.to_typename()}' not found."),
                    component_doc_tree,
                )
            )

    if validation_errors:
        for key, error, component_doc_tree in validation_errors:
            click.echo(
                error_dict_to_formatted_error(
                    key,
                    error,
                    source_position_tree=component_doc_tree.source_position_tree,
                    prefix=["attributes"] if key else [],
                )
            )
        return False
    else:
        click.echo("All component YAML validated successfully.")

        return True



import logging
import subprocess
from typing import Optional


def get_local_branch_name(project_dir: str) -> Optional[str]:
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                cwd=project_dir,
                stderr=subprocess.PIPE,
            )
            .decode("utf-8")
            .strip()
        )
    except subprocess.SubprocessError:
        logging.getLogger("dg").debug("Failed to determine git branch", exc_info=True)
        return None



import contextlib
import io
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import textwrap
from collections.abc import Iterator, Mapping, Sequence
from fnmatch import fnmatch
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional, TextIO, TypeVar, Union, overload

import click
from click_aliases import ClickAliasedGroup
from dagster_shared.utils import environ
from typing_extensions import Never, TypeAlias

from dagster_dg_core.error import DgError

if TYPE_CHECKING:
    import tomlkit

# There is some weirdness concerning the availabilty of hashlib.HASH between different Python
# versions, so for nowe we avoid trying to import it and just alias the type to Any.
Hash: TypeAlias = Any

CLI_CONFIG_KEY = "config"

DG_CLI_MAX_OUTPUT_WIDTH = 120


def show_dg_unlaunched_commands() -> bool:
    """We hide cli commands that we have not launched yet. Override this by setting
    the DG_SHOW_UNLAUNCHED_COMMANDS environment variable to any value.
    """
    return os.getenv("DG_SHOW_UNLAUNCHED_COMMANDS") is not None


def is_windows() -> bool:
    return sys.platform == "win32"


def is_macos() -> bool:
    return sys.platform == "darwin"


def is_uv_installed() -> bool:
    return shutil.which("uv") is not None


def get_activated_venv() -> Optional[Path]:
    """Returns the path to the activated virtual environment, or None if no virtual environment is active."""
    venv_path = os.environ.get("VIRTUAL_ENV")
    if venv_path:
        return Path(venv_path).absolute()
    return None


def resolve_local_venv(start_path: Path) -> Optional[Path]:
    path = start_path
    while path != path.parent:
        venv_path = path / ".venv"
        if venv_path.exists():
            return venv_path
        path = path.parent
    return None


def get_shortest_path_repr(abs_path: Path) -> Path:
    try:
        return abs_path.relative_to(Path.cwd())
    except ValueError:  # raised when path is not a descendant of cwd
        return abs_path


def get_logger(name: str, verbose: bool) -> logging.Logger:
    logger = logging.getLogger(name)
    log_level = logging.INFO if verbose else logging.WARNING
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter("%(message)s"))
    logger.handlers = [handler]
    logger.setLevel(log_level)
    logger.propagate = False
    return logger


def clear_screen():
    if is_windows():
        os.system("cls")
    else:
        os.system("clear")


def get_venv_activation_cmd(venv_dir: Path) -> str:
    if is_windows():
        return str(venv_dir / "Scripts" / "activate.bat")
    else:
        return f"source {venv_dir / 'bin' / 'activate'}"


def get_venv_executable(venv_dir: Path, executable: str = "python") -> Path:
    if is_windows():
        return venv_dir / "Scripts" / f"{executable}.exe"
    else:
        return venv_dir / "bin" / executable


def install_to_venv(venv_dir: Path, install_args: list[str]) -> None:
    executable = get_venv_executable(venv_dir)
    command = ["uv", "pip", "install", "--python", str(executable), *install_args]
    subprocess.run(command, check=True)


def is_valid_json(value: str) -> bool:
    try:
        json.loads(value)
        return True
    except json.JSONDecodeError:
        return False


# Short for "normalize path"-- use this to get the platform-correct string representation of an
# existing string path.
def cross_platfrom_string_path(path: str):
    return str(Path(path))


# uv commands should be executed in an environment with no pre-existing VIRTUAL_ENV set. If this
# variable is set (common during development) and does not match the venv resolved by uv, it prints
# undesireable warnings.
def strip_activated_venv_from_env_vars(env: Mapping[str, str]) -> Mapping[str, str]:
    return {k: v for k, v in env.items() if not k == "VIRTUAL_ENV"}


def discover_git_root(path: Path) -> Path:
    while path != path.parent:
        if (path / ".git").exists():
            return path
        path = path.parent
    raise ValueError("Could not find git root")


def discover_venv(path: Path) -> Path:
    while path != path.parent:
        if (path / ".venv").exists():
            return path
        path = path.parent
    raise ValueError("Could not find venv")


@contextlib.contextmanager
def pushd(path: Union[str, Path]) -> Iterator[None]:
    old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def camelcase(string: str) -> str:
    string = re.sub(r"^[\-_\.]", "", str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r"[\-_\.\s]([a-z])", lambda matched: str(matched.group(1)).upper(), string[1:]
    )


def snakecase(string: str) -> str:
    """Convert a string to snake_case.

    This function handles consecutive uppercase letters more gracefully than the naive approach.
    For example: "ACMEDatabricksJobComponent" -> "acme_databricks_job_component"
    Rather than: "ACMEDatabricksJobComponent" -> "a_c_m_e_databricks_job_component"
    """
    if not string:
        return string

    # First, handle sequences of uppercase letters followed by lowercase letters
    # This matches patterns like "ACME" in "ACMEDatabricks" -> "ACME_Databricks"
    # The pattern (?<![A-Z])([A-Z]+)(?=[A-Z][a-z]) matches:
    # - One or more uppercase letters ([A-Z]+)
    # - That are followed by an uppercase letter then lowercase ((?=[A-Z][a-z]))
    # - But not preceded by another uppercase letter ((?<![A-Z]))
    string = re.sub(r"([A-Z]+)(?=[A-Z][a-z])", r"\1_", string)

    # Add underscores before uppercase letters that follow lowercase letters or numbers
    # This handles the transition from lowercase/numbers to uppercase
    string = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", string)

    # Convert to lowercase
    string = string.lower()

    # Replace any non-alphanumeric characters with underscores
    string = re.sub(r"[^a-z0-9_]", "_", string)

    # Clean up multiple consecutive underscores
    string = re.sub(r"_+", "_", string)

    # Remove leading/trailing underscores
    string = string.strip("_")

    return string


_HELP_OUTPUT_GROUP_ATTR = "rich_help_panel"


# This is a private API of typer, but it is the only way I can see to get our global options grouped
# differently and it works well.
def set_option_help_output_group(param: click.Parameter, group: str) -> None:
    """Sets the help output group for a click parameter."""
    setattr(param, "rich_help_panel", group)


@contextlib.contextmanager
def modify_toml(path: Path) -> Iterator["tomlkit.TOMLDocument"]:
    import tomlkit

    toml = tomlkit.parse(path.read_text())
    yield toml
    path.write_text(tomlkit.dumps(toml))


@contextlib.contextmanager
def modify_toml_as_dict(path: Path) -> Iterator[dict[str, Any]]:  # unwrap gets the dict
    """Modify a TOML file as a plain python dict, destroying any comments or formatting.

    This is a destructive means of modifying TOML. We convert the parsed TOML document into a plain
    python object, modify it, and then write it back to the file. This will destroy comments and
    styling. It is useful mostly in a testing context where we want to e.g. set arbitrary invalid
    values in the file without worrying about the details of the TOML syntax (it has multiple kinds of
    dict-like objects, for instance).
    """
    import tomlkit

    toml_dict = load_toml_as_dict(path)
    yield toml_dict
    path.write_text(tomlkit.dumps(toml_dict))


def hash_directory_metadata(
    hasher: Hash,
    path: Union[str, Path],
    includes: Optional[Sequence[str]],
    excludes: Sequence[str],
    error_on_missing: bool,
) -> None:
    """Hashes the metadata of all files in the given directory.

    Args:
        hasher: The hasher to use to hash the metadata.
        path: The directory path to hash the metadata of.
        includes: The glob patterns of files to include in the hash, or None to include all files.
        excludes: The glob patterns of files to exclude from the hash.
        error_on_missing: Whether to raise an error if a file is missing. Set to False for cases where
            we expect the filesystem to be actively changing.
    """
    for root, dirs, files in os.walk(path):
        for name in dirs + files:
            if any(fnmatch(name, pattern) for pattern in excludes):
                continue
            if includes and not any(fnmatch(name, pattern) for pattern in includes):
                continue
            filepath = os.path.join(root, name)
            hash_file_metadata(hasher, filepath, error_on_missing)


def hash_file_metadata(hasher: Hash, path: Union[str, Path], error_on_missing) -> None:
    """Hashes the metadata of a file.

    Args:
        hasher: The hasher to use to hash the metadata.
        path: The file path to hash the metadata of.
        error_on_missing: Whether to raise an error if a file is missing.
    """
    try:
        stat = os.stat(path=path)
        hasher.update(str(path).encode())
        hasher.update(str(stat.st_mtime).encode())  # Last modified time
        hasher.update(str(stat.st_size).encode())  # File size
    except FileNotFoundError:
        if error_on_missing:
            raise


T = TypeVar("T")


def not_none(value: Optional[T]) -> T:
    if value is None:
        raise DgError("Expected non-none value.")
    return value


def exit_with_error(error_message: str, do_format: bool = True) -> Never:
    formatted_error_message = format_multiline_str(error_message) if do_format else error_message
    click.echo(click.style(formatted_error_message, fg="red"), err=True)
    sys.exit(1)


# ########################
# ##### ERROR MESSAGES
# ########################


def format_multiline_str(message: str) -> str:
    # width=10000 unwraps any hardwrapping
    dedented = textwrap.dedent(message).strip()
    paragraphs = [textwrap.fill(p, width=10000) for p in dedented.split("\n\n")]
    return "\n\n".join(paragraphs)


def generate_missing_registry_object_error_message(registry_object_key: str) -> str:
    return f"""
        No registry object `{registry_object_key}` is registered. Use `dg list components`
        to see the registered objects in your environment. You may need to install a package
        that provides `{registry_object_key}` into your environment.
    """


def generate_project_and_activated_venv_mismatch_warning(
    project_venv_path: Path,
    active_venv_path: Optional[Path],
) -> str:
    return f"""
        The active virtual environment does not match the virtual environment found in the project
        root directory. This may lead to unexpected behavior when running `dg` commands.

            active virtual environment: {active_venv_path}
            project virtual environment: {project_venv_path}
    """


NOT_WORKSPACE_ERROR_MESSAGE = """
This command must be run inside a Dagster workspace directory. Ensure that there is a
`pyproject.toml` file with `tool.dg.directory_type = "workspace"` set in the root workspace
directory.
"""


NOT_PROJECT_ERROR_MESSAGE = """
This command must be run inside a Dagster project directory. Ensure that the nearest
pyproject.toml has `tool.dg.directory_type = "project"` set.
"""

NOT_WORKSPACE_OR_PROJECT_ERROR_MESSAGE = """
This command must be run inside a Dagster workspace or project directory. Ensure that the
nearest pyproject.toml has `tool.dg.directory_type = "project"` or `tool.dg.directory_type =
"workspace"` set.
"""


def msg_with_potential_paths(message: str, potential_paths: list[Path]) -> str:
    paths_str = "\n".join([f"- {p}" for p in potential_paths])
    return f"""{message}
You may have wanted to run this command in the following directory:

{paths_str}
"""


NOT_COMPONENT_LIBRARY_ERROR_MESSAGE = """
This command must be run inside a Dagster component library directory. Ensure that the nearest
pyproject.toml has an entry point defined under the `dagster_dg_cli.registry_modules` group.
"""


def generate_tool_dg_cli_in_project_in_workspace_error_message(
    project_path: Path, workspace_path: Path
) -> str:
    return textwrap.dedent(f"""
        The `tool.dg.cli` section is ignored for project `pyproject.toml` files inside of a
        workspace. Any `tool.dg.cli` settings should be placed in the workspace config file.

        `cli` section detected in workspace project `pyproject.toml` file at:
            {project_path}
        """).strip()


# ########################
# ##### CUSTOM CLICK SUBCLASSES
# ########################

# Here we subclass click.Command and click.Group to customize the help output. We do this in order
# to visually separate global from command-specific options. The form of the output can be seen in
# dagster_dg_core_tests.test_custom_help_format.


class DgClickHelpMixin:
    def format_help(self, context: click.Context, formatter: click.HelpFormatter):
        """Customizes the help to include hierarchical usage."""
        from typer.rich_utils import rich_format_help

        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")

        # We use typer's rich_format_help to render our help output, despite the fact that we are
        # not using typer elsewhere in the app. Global options are separated from command-specific
        # options by setting the `rich_help_panel` attribute to "Global options" on our global options.
        rich_format_help(obj=self, ctx=context, markup_mode="rich")


class DgClickCommand(DgClickHelpMixin, click.Command):  # pyright: ignore[reportIncompatibleMethodOverride]
    def __init__(self, *args, unlaunched: bool = False, **kwargs):
        """DgClickCommand with conditional hiding for unlaunched features.

        Args:
            unlaunched: If True, the command will be hidden unless DG_SHOW_UNLAUNCHED_COMMANDS
                environment variable is set.
        """
        if unlaunched:
            kwargs["hidden"] = not show_dg_unlaunched_commands()
        super().__init__(*args, **kwargs)


class DgClickGroup(DgClickHelpMixin, ClickAliasedGroup):  # pyright: ignore[reportIncompatibleMethodOverride]
    def __init__(self, *args, unlaunched: bool = False, **kwargs):
        """DgClickGroup with conditional hiding for unlaunched features.

        Args:
            unlaunched: If True, the group will be hidden unless DG_SHOW_UNLAUNCHED_COMMANDS
                environment variable is set.
        """
        if unlaunched:
            kwargs["hidden"] = not show_dg_unlaunched_commands()
        super().__init__(*args, **kwargs)


# ########################
# ##### JSON SCHEMA
# ########################

_JSON_SCHEMA_TYPE_TO_CLICK_TYPE = {"string": str, "integer": int, "number": float, "boolean": bool}


def _get_field_type_info_from_field_info(field_info: Mapping[str, Any]) -> Mapping[str, Any]:
    """Extract the dict holding field type info (in particular, the type and whether it is an array)
    from a JSON schema field info dict.

    If the field info is not a union type, returns itself.
    If the field info is a union type, returns the first non-null type.
    If the field info has no type info, default to type info for type "string".
    """
    if field_info.get("type"):
        return field_info
    else:
        return next(
            (t for t in field_info.get("anyOf", []) if t.get("type") not in (None, "null")),
            {"type": "string"},
        )


def json_schema_property_to_click_option(
    key: str, field_info: Mapping[str, Any], required: bool
) -> click.Option:
    # Extract the dict holding field type info (in particular, the type and whether it is an array)
    # This might be nested in an anyOf block
    field_type_info = _get_field_type_info_from_field_info(field_info)
    is_array_type = field_type_info.get("type") == "array"
    field_type = (
        field_type_info.get("items", {}).get("type") or field_type_info.get("type") or "string"
    )

    option_name = f"--{key.replace('_', '-')}"

    # Handle object type fields as JSON strings
    if field_type == "object":
        option_type = str  # JSON string input
        help_text = f"[scaffolder parameter] {key} (JSON string)"
        callback = parse_json_option

    # Handle other basic types
    else:
        option_type = _JSON_SCHEMA_TYPE_TO_CLICK_TYPE[field_type]
        help_text = f"(scaffolder param) {key}"
        callback = None

    return click.Option(
        [option_name],
        type=option_type,
        required=required,
        help=help_text,
        callback=callback,
        multiple=is_array_type,
    )


def parse_json_option(context: click.Context, param: click.Option, value: str):
    """Callback to parse JSON string options into Python objects."""
    if value:
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            raise click.BadParameter(f"Invalid JSON string for '{param.name}'.")
    return value


# ########################
# ##### TOML MANIPULATION
# ########################

TomlPath: TypeAlias = tuple[Union[str, int], ...]
TomlDoc: TypeAlias = Union["tomlkit.TOMLDocument", dict[str, Any]]


def load_toml_as_dict(path: Path) -> dict[str, Any]:
    import tomlkit

    return tomlkit.parse(path.read_text()).unwrap()


def get_toml_node(
    doc: TomlDoc,
    path: TomlPath,
    expected_type: Union[type[T], tuple[type[T], ...]],
) -> T:
    """Given a tomlkit-parsed document/table (`doc`),retrieve the nested value at `path` and ensure
    it is of type `expected_type`. Returns the value if so, or raises a KeyError / TypeError if not.
    """
    value = _gather_toml_nodes(doc, path)[-1]
    if not isinstance(value, expected_type):
        expected_types = expected_type if isinstance(expected_type, tuple) else (expected_type,)
        type_str = " or ".join(t.__name__ for t in expected_types)
        raise TypeError(
            f"Expected '{toml_path_to_str(path)}' to be {type_str}, "
            f"but got {type(value).__name__} instead."
        )
    return value


def has_toml_node(doc: TomlDoc, path: TomlPath) -> bool:
    """Given a tomlkit-parsed document/table (`doc`), return whether a value is defined at `path`."""
    result = _gather_toml_nodes(doc, path, error_on_missing=False)
    return False if result is None else True


def delete_toml_node(doc: TomlDoc, path: TomlPath) -> None:
    """Given a tomlkit-parsed document/table (`doc`), delete the nested value at `path`. Raises
    an error if the leading keys do not already lead to a TOML container node.
    """
    import tomlkit

    nodes = _gather_toml_nodes(doc, path)
    container = nodes[-2] if len(nodes) > 1 else doc
    key_or_index = path[-1]
    if isinstance(container, dict):
        assert isinstance(key_or_index, str)  # We already know this from _traverse_toml_path
        del container[key_or_index]
    elif isinstance(container, tomlkit.TOMLDocument):
        assert isinstance(key_or_index, str)  # We already know this from _traverse_toml_path
        container.remove(key_or_index)
    elif isinstance(container, list):
        assert isinstance(key_or_index, int)  # We already know this from _traverse_toml_path
        container.pop(key_or_index)
    else:
        raise Exception("Unreachable.")


def set_toml_node(doc: TomlDoc, path: TomlPath, value: object) -> None:
    """Given a tomlkit-parsed document/table (`doc`),set a nested value at `path` to `value`. Raises
    an error if the leading keys do not already lead to a TOML container node.
    """
    container = _gather_toml_nodes(doc, path[:-1])[-1] if len(path) > 1 else doc
    key_or_index = path[-1]  # type: ignore  # pyright bug
    if isinstance(container, dict):
        if not isinstance(key_or_index, str):
            raise TypeError(f"Expected key to be a string, but got {type(key_or_index).__name__}")
        container[key_or_index] = value
    elif isinstance(container, list):
        if not isinstance(key_or_index, int):
            raise TypeError(f"Expected key to be an integer, but got {type(key_or_index).__name__}")
        container[key_or_index] = value
    else:
        raise Exception("Unreachable.")


@overload
def _gather_toml_nodes(
    doc: TomlDoc, path: TomlPath, error_on_missing: Literal[True] = ...
) -> list[Any]: ...


@overload
def _gather_toml_nodes(
    doc: TomlDoc, path: TomlPath, error_on_missing: Literal[False] = ...
) -> Optional[list[Any]]: ...


def _gather_toml_nodes(
    doc: TomlDoc, path: TomlPath, error_on_missing: bool = True
) -> Optional[list[Any]]:
    nodes: list[Any] = []
    current: Any = doc
    for key in path:
        if isinstance(key, str):
            if not isinstance(current, dict) or key not in current:
                if error_on_missing:
                    raise KeyError(f"Key '{key}' not found in path: {toml_path_to_str(path)}")
                return None
            current = current[key]
        elif isinstance(key, int):
            if not isinstance(current, list) or key < 0 or key >= len(current):
                if error_on_missing:
                    raise KeyError(f"Index '{key}' not found in path: {toml_path_to_str(path)}")
                return None
            current = current[key]
        else:
            raise TypeError(f"Expected key to be a string or integer, but got {type(key)}")
        nodes.append(current)

    return nodes


def toml_path_to_str(path: TomlPath) -> str:
    if len(path) == 0:
        return ""
    first = path[0]
    if not isinstance(first, str):
        raise TypeError(f"Expected first element of path to be a string, but got {type(first)}")
    elif len(path) == 1:
        return first
    else:
        str_path = first
        for item in path[1:]:
            if isinstance(item, int):
                str_path += f"[{item}]"
            elif isinstance(item, str):
                str_path += f".{item}"
            else:
                raise TypeError(
                    f"Expected path elements to be strings or integers, but got {type(item)}"
                )
    return str_path


def toml_path_from_str(path: str) -> TomlPath:
    tokens = []
    for segment in path.split("."):
        # Split each segment by bracketed chunks, e.g. "key[1]" -> ["key", "[1]"]
        parts = re.split(r"(\[\d+\])", segment)
        for p in parts:
            if not p:  # Skip empty strings
                continue
            if p.startswith("[") and p.endswith("]"):
                tokens.append(int(p[1:-1]))  # Convert "[1]" to integer 1
            else:
                tokens.append(p)
    return tuple(tokens)


def create_toml_node(
    doc: dict[str, Any],
    path: tuple[Union[str, int], ...],
    value: object,
) -> None:
    """Set a toml node at a path that consists of a sequence of keys and integer indices.
    Intermediate containers that don't yet exist will be created along the way based on the types of
    the keys. Note that this does not support TOMLDocument objects, only plain dictionaries. The
    reason is that the correct type of container to insert at intermediate nodes is ambiguous for
    TOMLDocmuent objects.
    """
    import tomlkit

    if isinstance(doc, tomlkit.TOMLDocument):
        raise TypeError(
            "`create_toml_node` only works on the plain dictionary representation of a TOML document."
        )
    current: Any = doc
    for i, key in enumerate(path):
        is_final_key = i == len(path) - 1
        if isinstance(key, str):
            if not isinstance(current, dict):
                raise KeyError(f"Key '{key}' not found in path: {toml_path_to_str(path)}")
            elif is_final_key:
                current[key] = value
            elif key not in current:
                current[key] = _get_new_container_node(path[i + 1])
            current = current[key]
        elif isinstance(key, int):
            if not isinstance(current, list):
                raise KeyError(f"Index '{key}' not found in path: {toml_path_to_str(path)}")
            is_key_in_range = key >= 0 and key < len(current)
            is_append_key = key == len(current)
            if is_final_key and is_key_in_range:
                current[key] = value
            elif is_final_key and is_append_key:
                current.append(value)
            elif is_key_in_range:
                current = current[key]
            elif is_append_key:
                current.append(_get_new_container_node(path[i + 1]))
                current = current[key]
            else:
                raise KeyError(f"Key '{key}' not found in path: {toml_path_to_str(path)}")
        else:
            raise TypeError(f"Expected key to be a string or integer, but got {type(key)}")


def _get_new_container_node(
    representative_key: Union[int, str],
) -> Union[dict[str, Any], list[Any]]:
    return [] if isinstance(representative_key, int) else {}


@contextlib.contextmanager
def capture_stdout() -> Iterator[TextIO]:
    """Capture stdout and return it as a string."""
    stdout = sys.stdout
    string_buffer = io.StringIO()
    try:
        sys.stdout = string_buffer
        yield string_buffer
    finally:
        sys.stdout = stdout


@contextlib.contextmanager
def activate_venv(venv_path: Union[str, Path]) -> Iterator[None]:
    """Simulated activation of the passed in virtual environment for the current process."""
    venv_path = (Path(venv_path) if isinstance(venv_path, str) else venv_path).absolute()
    with environ(
        {
            "VIRTUAL_ENV": str(venv_path),
            "PATH": os.pathsep.join(
                [
                    str(venv_path / ("Scripts" if sys.platform == "win32" else "bin")),
                    os.getenv("PATH", ""),
                ]
            ),
        }
    ):
        yield



import hashlib
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from dagster_shared.scaffold import DEFAULT_FILE_EXCLUDE_PATTERNS

from dagster_dg_core.utils import hash_directory_metadata, hash_file_metadata


def hash_paths(
    paths: Sequence[Path],
    includes: Optional[Sequence[str]] = None,
    excludes: Sequence[str] = DEFAULT_FILE_EXCLUDE_PATTERNS,
    error_on_missing: bool = True,
) -> str:
    """Hash the given paths, including their metadata.

    Args:
        paths: The paths to hash.
        includes: A list of glob patterns to target, excluding files that don't match any of the patterns.
        excludes: A list of glob patterns to exclude, including files that match any of the patterns. Defaults to
            various Python-related files that are not relevant to the contents of the project.

    Returns:
        The hash of the paths.
    """
    hasher = hashlib.md5()
    for path in paths:
        if path.is_dir():
            hash_directory_metadata(
                hasher,
                path,
                includes=includes,
                excludes=excludes,
                error_on_missing=error_on_missing,
            )
        else:
            hash_file_metadata(hasher, path, error_on_missing=error_on_missing)
    return hasher.hexdigest()



from collections import defaultdict
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional

from dagster_shared import check
from typing_extensions import Self

from dagster_dg_core.context import DgContext


def get_specified_env_var_deps(component_data: Mapping[str, Any]) -> set[str]:
    if "requirements" not in component_data or "env" not in component_data["requirements"]:
        return set()
    return set(component_data["requirements"]["env"])


def get_project_specified_env_vars(dg_context: DgContext) -> Mapping[str, Sequence[Path]]:
    """Returns a mapping of environment variables to the components that specify
    requiring them.
    """
    if not dg_context.has_defs_path:
        return {}

    from dagster_shared.yaml_utils import parse_yamls_with_source_position
    from yaml.scanner import ScannerError

    env_vars = defaultdict(list)

    for component_dir in dg_context.defs_path.rglob("*"):
        defs_path = component_dir / "defs.yaml"

        if defs_path.exists():
            text = defs_path.read_text()
            try:
                component_doc_trees = parse_yamls_with_source_position(
                    text, filename=str(defs_path)
                )
            except ScannerError:
                continue

            for component_doc_tree in component_doc_trees:
                specified_env_var_deps = get_specified_env_var_deps(component_doc_tree.value)
                for key in specified_env_var_deps:
                    env_vars[key].append(defs_path.relative_to(dg_context.defs_path).parent)
    return env_vars


class ProjectEnvVars:
    """Represents the environment for a project, stored in the .env file of a
    project root.
    """

    def __init__(self, ctx: DgContext, values: Mapping[str, Optional[str]]):
        self.ctx = ctx
        self._values = values

    @classmethod
    def empty(cls, ctx: DgContext) -> "Self":
        return cls(ctx, values={})

    @classmethod
    def from_ctx(cls, ctx: DgContext) -> "Self":
        check.invariant(ctx.is_project, "ProjectEnvVars can only be created from a project context")
        from dotenv import dotenv_values

        env_path = ctx.root_path / ".env"
        if not env_path.exists():
            return cls(ctx, values={})
        env = dotenv_values(env_path)
        return cls(ctx, values=env)

    @property
    def values(self) -> Mapping[str, Optional[str]]:
        return self._values

    def get(self, key: str) -> Optional[str]:
        return self.values.get(key)

    def with_values(self, values: Mapping[str, Optional[str]]) -> "ProjectEnvVars":
        return ProjectEnvVars(self.ctx, {**self.values, **values})

    def without_values(self, keys: set[str]) -> "ProjectEnvVars":
        return ProjectEnvVars(self.ctx, {k: v for k, v in self.values.items() if k not in keys})

    def write(self) -> None:
        env_path = self.ctx.root_path / ".env"
        env_path.write_text(
            "\n".join([f"{key}={value}" for key, value in self.values.items() if value is not None])
        )



import datetime
import logging
import re
from collections.abc import Iterable, Mapping
from functools import cached_property
from pathlib import Path
from typing import Any, Final, Optional

import dagster_shared.check as check
from dagster_shared.record import record
from dagster_shared.serdes.serdes import whitelist_for_serdes
from dagster_shared.seven import resolve_module_pattern
from dagster_shared.utils.config import get_canonical_defs_module_name
from packaging.version import Version
from typing_extensions import Self

from dagster_dg_core.component import EnvRegistry
from dagster_dg_core.config import (
    DgConfig,
    DgRawBuildConfig,
    DgRawCliConfig,
    DgWorkspaceProjectSpec,
    discover_and_validate_config_files,
    is_workspace_file_config,
    modify_dg_toml_config,
    raise_file_config_validation_error,
)
from dagster_dg_core.error import DgError
from dagster_dg_core.utils import (
    NOT_COMPONENT_LIBRARY_ERROR_MESSAGE,
    NOT_PROJECT_ERROR_MESSAGE,
    NOT_WORKSPACE_ERROR_MESSAGE,
    NOT_WORKSPACE_OR_PROJECT_ERROR_MESSAGE,
    exit_with_error,
    generate_project_and_activated_venv_mismatch_warning,
    get_activated_venv,
    get_logger,
    get_toml_node,
    get_venv_executable,
    has_toml_node,
    msg_with_potential_paths,
    set_toml_node,
)
from dagster_dg_core.utils.warnings import emit_warning

# Project
_DEFAULT_PROJECT_CODE_LOCATION_TARGET_MODULE: Final = "definitions"
_DEFAULT_PROJECT_PLUGIN_MODULE: Final = "components"
_DEFAULT_PROJECT_PLUGIN_MODULE_REGISTRY_FILE: Final = "plugin_modules.json"
_EXCLUDED_COMPONENT_DIRECTORIES: Final = {"__pycache__"}
DG_PLUGIN_ENTRY_POINT_GROUP: Final = "dagster_dg_cli.registry_modules"
# Remove in future, in place for backcompat
OLD_DG_PLUGIN_ENTRY_POINT_GROUPS = [
    "dagster_dg.library",
    "dagster_dg.plugin",
    "dagster_dg_cli.plugin",
]


def _should_capture_components_cli_stderr() -> bool:
    """Used in tests to pass along stderr from the components CLI to the parent process."""
    return False


class DgContext:
    root_path: Path
    config: DgConfig
    cli_opts: Optional[DgRawCliConfig] = None
    _workspace_root_path: Optional[Path]

    # We need to preserve CLI options for the context to be able to derive new contexts, because
    # cli_options override everything else. If we didn't maintain them we wouldn't be able to tell
    # whether a given config option should be overridden in a new derived context.
    def __init__(
        self,
        config: DgConfig,
        root_path: Path,
        workspace_root_path: Optional[Path] = None,
        cli_opts: Optional[DgRawCliConfig] = None,
    ):
        self.config = config
        self.root_path = root_path
        self._workspace_root_path = workspace_root_path
        self.cli_opts = cli_opts
        self.component_registry = EnvRegistry.empty()

        # Always run this check, its a no-op if there is no pyproject.toml.
        _validate_plugin_entry_point(self)

    @classmethod
    def for_workspace_environment(cls, path: Path, command_line_config: DgRawCliConfig) -> Self:
        context = cls.from_file_discovery_and_command_line_config(path, command_line_config)

        # Commands that operate on a workspace need to be run inside a workspace context.
        if not context.is_in_workspace:
            potential_paths = cls._locate_potential_projects_or_workspaces(
                path, command_line_config, allow_projects=False, allow_workspaces=True
            )
            if potential_paths:
                exit_with_error(
                    msg_with_potential_paths(NOT_WORKSPACE_ERROR_MESSAGE, potential_paths)
                )
            exit_with_error(NOT_WORKSPACE_ERROR_MESSAGE)
        return context

    @classmethod
    def for_project_environment(cls, path: Path, command_line_config: DgRawCliConfig) -> Self:
        context = cls.from_file_discovery_and_command_line_config(path, command_line_config)

        if not context.is_project:
            potential_paths = cls._locate_potential_projects_or_workspaces(
                path, command_line_config, allow_projects=True, allow_workspaces=False
            )
            if potential_paths:
                exit_with_error(
                    msg_with_potential_paths(NOT_PROJECT_ERROR_MESSAGE, potential_paths)
                )
            exit_with_error(NOT_PROJECT_ERROR_MESSAGE)
        _validate_project_venv_activated(context)
        return context

    @classmethod
    def _locate_potential_projects_or_workspaces(
        cls,
        path: Path,
        command_line_config: DgRawCliConfig,
        allow_projects: bool,
        allow_workspaces: bool,
    ) -> list[Path]:
        """Locates potential project or workspace directories by iterating over parent directories
        or immediate children, to present to the user if they run a command in a directory that
        is not a project or workspace.
        """
        potential_paths = [
            *path.parents,
            *path.iterdir(),
        ]
        matching_paths = []
        for path in potential_paths:
            context = cls.from_file_discovery_and_command_line_config(path, command_line_config)
            if context.is_in_workspace and allow_workspaces:
                matching_paths.append(path)
            elif context.is_project and allow_projects:
                matching_paths.append(path)
        return matching_paths

    @classmethod
    def for_workspace_or_project_environment(
        cls, path: Path, commmand_line_config: DgRawCliConfig
    ) -> Self:
        context = cls.from_file_discovery_and_command_line_config(path, commmand_line_config)

        # Commands that operate on a workspace need to be run inside a workspace or project
        # context.
        if not (context.is_in_workspace or context.is_project):
            potential_paths = cls._locate_potential_projects_or_workspaces(
                path, commmand_line_config, allow_projects=True, allow_workspaces=True
            )
            if potential_paths:
                exit_with_error(
                    msg_with_potential_paths(
                        NOT_WORKSPACE_OR_PROJECT_ERROR_MESSAGE, potential_paths
                    )
                )
            exit_with_error(NOT_WORKSPACE_OR_PROJECT_ERROR_MESSAGE)
        if context.is_project:
            _validate_project_venv_activated(context)
        return context

    @classmethod
    def for_component_library_environment(
        cls, path: Path, command_line_config: DgRawCliConfig
    ) -> Self:
        context = cls.from_file_discovery_and_command_line_config(path, command_line_config)

        # Commands that operate on a component library need to be run in a component library context.

        if not context.has_registry_module_entry_point and not context.is_project:
            exit_with_error(NOT_COMPONENT_LIBRARY_ERROR_MESSAGE)
        return context

    @classmethod
    def from_file_discovery_and_command_line_config(
        cls,
        path: Path,
        command_line_config: DgRawCliConfig,
        *,
        emit_log: bool = False,
    ) -> Self:
        result = discover_and_validate_config_files(path)

        if result.has_root_file and result.root_result.has_errors:
            raise_file_config_validation_error(result.root_result.message, path)
        elif result.has_container_workspace_file:
            if result.container_workspace_result.has_errors:
                raise_file_config_validation_error(
                    result.container_workspace_result.message,
                    check.not_none(result.container_workspace_file_path),
                )
            elif not is_workspace_file_config(result.container_workspace_result.config):
                raise_file_config_validation_error("Expected a workspace configuration.", path)

        config = DgConfig.from_partial_configs(
            root_file_config=result.root_config,
            container_workspace_file_config=result.container_workspace_config,
            command_line_config=command_line_config,
            user_config=result.user_config,
        )
        if result.cli_config_warning:
            emit_warning(
                "cli_config_in_workspace_project",
                result.cli_config_warning,
                config.cli.suppress_warnings,
            )

        return cls(
            config=config,
            root_path=result.root_path,
            workspace_root_path=result.workspace_root_path,
            cli_opts=command_line_config,
        )

    @classmethod
    def default(cls) -> Self:
        return cls(DgConfig.default(), Path.cwd())

    # Use to derive a new context for a project while preserving existing settings
    def with_root_path(self, root_path: Path) -> Self:
        if not ((root_path / "pyproject.toml").exists() or (root_path / "dg.toml").exists()):
            raise DgError(f"Cannot find `pyproject.toml` at {root_path}")
        return self.__class__.from_file_discovery_and_command_line_config(
            root_path, self.cli_opts or {}
        )

    def component_registry_paths(self) -> list[Path]:
        """Paths that should be watched for changes to the component registry."""
        return [
            self.root_path / "uv.lock",
            *([self.default_registry_module_path] if self.has_registry_module_entry_point else []),
        ]

    # ########################
    # ##### WORKSPACE METHODS
    # ########################

    @property
    def is_in_workspace(self) -> bool:
        return self._workspace_root_path is not None

    @property
    def workspace_root_path(self) -> Path:
        if not self._workspace_root_path:
            raise DgError("`workspace_root_path` is only available in a workspace context")
        return self._workspace_root_path

    def has_project(self, relative_path: Path) -> bool:
        if not self.is_in_workspace:
            raise DgError("`has_project` is only available within a workspace")
        return bool(
            next(
                (spec for spec in self.project_specs if spec.path == relative_path),
                None,
            )
        )

    @property
    def project_specs(self) -> list[DgWorkspaceProjectSpec]:
        if not self.config.workspace:
            raise DgError("`project_specs` is only available within a workspace")
        return self.config.workspace.projects

    # ########################
    # ##### GENERAL PYTHON PACKAGE METHODS
    # ########################

    @property
    def root_module_name(self) -> str:
        if self.config.project:
            return self.config.project.root_module
        elif self.has_registry_module_entry_point:
            return self.default_registry_root_module_name.split(".")[0]
        else:
            raise DgError("Cannot determine root package name")

    @property
    def root_module_path(self) -> Path:
        return self.get_path_for_local_module(self.root_module_name)

    # ########################
    # ##### PROJECT METHODS
    # ########################

    @property
    def is_project(self) -> bool:
        return self.config.project is not None

    @property
    def project_name(self) -> str:
        if not self.is_project:
            raise DgError("`project_name` is only available in a Dagster project context")
        return self.root_path.name

    @property
    def project_python_executable(self) -> Path:
        if not self.is_project:
            raise DgError(
                "`project_python_executable` is only available in a Dagster project context"
            )
        return self.root_path / get_venv_executable(Path(".venv"))

    @cached_property
    def build_config_path(self) -> Path:
        return self.root_path / "build.yaml"

    @cached_property
    def build_config(self) -> Optional[DgRawBuildConfig]:
        import yaml

        build_yaml_path = self.build_config_path

        if not build_yaml_path.resolve().exists():
            return None

        with open(build_yaml_path) as f:
            build_config_dict = yaml.safe_load(f)
            build_directory = build_config_dict.get("directory")
            if build_directory:
                build_directory_path = Path(build_directory)
                if not build_directory_path.is_absolute():
                    build_directory_path = (build_yaml_path.parent / build_directory_path).resolve()
                build_config_dict["directory"] = str(build_directory_path.resolve())

            return build_config_dict

    @cached_property
    def container_context_config_path(self) -> Path:
        return self.root_path / "container_context.yaml"

    @cached_property
    def container_context_config(self) -> Optional[Mapping[str, Any]]:
        import yaml

        container_context_yaml_path = self.container_context_config_path

        if not container_context_yaml_path.resolve().exists():
            return None

        with open(container_context_yaml_path) as f:
            return yaml.safe_load(f)

    @cached_property
    def defs_module_name(self) -> str:
        if not self.config.project:
            raise DgError("`defs_module_name` is only available in a Dagster project context")
        return get_canonical_defs_module_name(
            self.config.project.defs_module, self.root_module_name
        )

    @cached_property
    def _defs_path(self) -> Path:
        defs_module_name = self.defs_module_name
        return self.get_path_for_local_module(defs_module_name, require_exists=False)

    @cached_property
    def has_defs_path(self) -> bool:
        return self._defs_path.exists()

    @property
    def defs_path(self) -> Path:
        if not self.has_defs_path:
            raise DgError(
                f"Defs folder not found. Ensure folder `{self._defs_path.relative_to(self.root_path)}` exists in the project root."
            )
        return self._defs_path

    def get_component_instance_names(self) -> Iterable[str]:
        return [
            str(p.name)
            for p in self.defs_path.iterdir()
            if p.is_dir() and str(p.name) not in _EXCLUDED_COMPONENT_DIRECTORIES
        ]

    def get_component_instance_module_name(self, name: str) -> str:
        return f"{self.defs_module_name}.{name}"

    def has_object_at_defs_path(self, defs_path: str) -> bool:
        return (self.defs_path / defs_path).exists()

    @property
    def target_args(self) -> Mapping[str, str]:
        if not self.config.project:
            raise DgError("`target_args` are only available in a Dagster project context")

        return {"module_name": self.code_location_target_module_name}

    @property
    def code_location_target_module_name(self) -> str:
        if not self.config.project:
            raise DgError(
                "`code_location_target_module_name` is only available in a Dagster project context"
            )
        return (
            self.config.project.code_location_target_module
            or f"{self.root_module_name}.{_DEFAULT_PROJECT_CODE_LOCATION_TARGET_MODULE}"
        )

    @cached_property
    def code_location_target_path(self) -> Path:
        return self.get_path_for_local_module(self.code_location_target_module_name)

    @property
    def code_location_name(self) -> str:
        if not self.config.project:
            raise DgError("`code_location_name` is only available in a Dagster project context")
        return self.config.project.code_location_name or self.project_name

    # ########################
    # ##### PLUGIN METHODS
    # ########################

    # It is possible for a single package to define multiple entry points under the
    # `dagster_dg_cli.registry_modules` entry point group. At present, `dg` only cares about the first one, which
    # it uses for all component type scaffolding operations.

    @property
    def has_registry_module_entry_point(self) -> bool:
        return bool(self._dagster_components_entry_points)

    @cached_property
    def default_registry_root_module_name(self) -> str:
        if self._dagster_components_entry_points:
            return next(iter(self._dagster_components_entry_points.values()))
        elif self.is_project:
            return f"{self.root_module_name}.{_DEFAULT_PROJECT_PLUGIN_MODULE}"
        else:
            raise DgError(
                "`default_component_library_module_name` is only available in a component library context"
            )

    @cached_property
    def default_registry_module_path(self) -> Path:
        if not self.has_registry_module_entry_point and not self.is_project:
            raise DgError(
                "`default_plugin_module_path` is only available in a component library context"
            )
        return self.get_path_for_local_module(
            self.default_registry_root_module_name, require_exists=False
        )

    @cached_property
    def project_registry_modules(self) -> list[str]:
        """Return a list of resolved plugin module references for the current project.

        This resolves any wildcard patterns in the raw registry_modules configuration
        into actual module names.
        """
        if not self.config.project:
            raise DgError(
                "`project_registry_modules` is only available in a Dagster project context"
            )
        return sorted(
            set(
                mod
                for pattern in self.config.project.registry_modules
                for mod in resolve_module_pattern(pattern)
            )
        )

    def add_project_registry_module(self, module_name: str) -> None:
        """Add a module name to the project plugin module registry."""
        import tomlkit
        import tomlkit.items

        if not self.is_project:
            raise DgError(
                "`add_project_registry_module` is only available in a Dagster project context"
            )
        with modify_dg_toml_config(self.config_file_path) as toml:
            if has_toml_node(toml, ("project", "registry_modules")):
                registry_modules = get_toml_node(
                    toml, ("project", "registry_modules"), tomlkit.items.Array
                )
                registry_modules.add_line(module_name)
            else:
                set_toml_node(toml, ("project", "registry_modules"), tomlkit.array())
                registry_modules = get_toml_node(
                    toml, ("project", "registry_modules"), tomlkit.items.Array
                )
                registry_modules.add_line(module_name)
                registry_modules.add_line(indent="")

    @cached_property
    def _dagster_components_entry_points(self) -> Mapping[str, str]:
        # defered imports for perf
        import tomlkit
        import tomlkit.items

        if self.pyproject_toml_path.exists():
            toml = tomlkit.parse(self.pyproject_toml_path.read_text())
            if has_toml_node(toml, ("project", "entry-points", DG_PLUGIN_ENTRY_POINT_GROUP)):
                return get_toml_node(
                    toml,
                    ("project", "entry-points", DG_PLUGIN_ENTRY_POINT_GROUP),
                    (tomlkit.items.Table, tomlkit.items.InlineTable),
                ).unwrap()
            # Keeping for backwards compatibility. Should be removed eventually.
            for entry_point_group in OLD_DG_PLUGIN_ENTRY_POINT_GROUPS:
                if has_toml_node(toml, ("project", "entry-points", entry_point_group)):
                    return get_toml_node(
                        toml,
                        ("project", "entry-points", entry_point_group),
                        (tomlkit.items.Table, tomlkit.items.InlineTable),
                    ).unwrap()
        if self.setup_cfg_path.exists():
            import warnings

            from setuptools import SetuptoolsDeprecationWarning
            from setuptools.config import read_configuration

            # Ignore deprecation warnings for the read_configuration function
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=SetuptoolsDeprecationWarning)
                config = read_configuration("setup.cfg")
            entry_points = config.get("options", {}).get("entry_points", {})
            group = next(
                (
                    group
                    for group in (DG_PLUGIN_ENTRY_POINT_GROUP, *OLD_DG_PLUGIN_ENTRY_POINT_GROUPS)
                    if group in entry_points
                ),
                None,
            )
            raw_plugin_entry_points = entry_points[group] if group else []
            plugin_entry_points = {}
            for entry_point in raw_plugin_entry_points:
                k, v = re.split(r"\s*=\s*", entry_point, 1)
                plugin_entry_points[k] = v
            return plugin_entry_points
        return {}

    # ########################
    # ##### LOG METHODS
    # ########################

    @cached_property
    def log(self) -> logging.Logger:
        """Return a logger for this context.

        The logger is configured based on the verbose setting in config.
        Default level is WARNING, and if verbose is set, level is increased to INFO.

        Returns:
            A configured logger that can be used with methods like info(), debug(), warning(), etc.
        """
        return get_logger("dagster_dg.context", self.config.cli.verbose)

    # ########################
    # ##### HELPERS
    # ########################

    @property
    def config_file_path(self) -> Path:
        return self.dg_toml_path if self.dg_toml_path.exists() else self.pyproject_toml_path

    @property
    def dg_toml_path(self) -> Path:
        return self.root_path / "dg.toml"

    @property
    def pyproject_toml_path(self) -> Path:
        return self.root_path / "pyproject.toml"

    @property
    def setup_cfg_path(self) -> Path:
        return self.root_path / "setup.cfg"

    def get_path_for_local_module(self, module_name: str, require_exists: bool = True) -> Path:
        if not self.is_project and not self.has_registry_module_entry_point:
            raise DgError(
                "`get_path_for_local_module` is only available in a project or component library context"
            )
        if not module_name.startswith(self.root_module_name):
            raise DgError(f"Module `{module_name}` is not part of the current project.")

        # Attempt to resolve the path for a local module by looking in both `src` and the root
        # level. Unfortunately there is no setting reliably present in pyproject.toml or setup.py
        # that can be relied on to know in advance the package root (src or root level).

        base_path = self.root_path / "src" if (self.root_path / "src").exists() else self.root_path
        path = base_path / Path(*module_name.split("."))
        if path.with_suffix(".py").exists():
            return path.with_suffix(".py")

        # Note that when `require_exists` is False, the returned path assumes a directory
        # instead of a py file.
        elif path.exists() or not require_exists:
            return path

        raise DgError(f"Cannot find module `{module_name}` in the current project.")


# ########################
# ##### HELPERS
# ########################


def _validate_project_venv_activated(context: DgContext) -> None:
    if not context.config.project:
        raise DgError(
            "`_validate_project_venv_activated` is only available in a Dagster project context"
        )
    activated_venv = get_activated_venv()
    project_venv = context.root_path / ".venv"
    if project_venv.exists() and (
        not activated_venv or project_venv.resolve() != activated_venv.resolve()
    ):
        msg = generate_project_and_activated_venv_mismatch_warning(project_venv, activated_venv)
        emit_warning(
            "project_and_activated_venv_mismatch",
            msg,
            context.config.cli.suppress_warnings,
        )


# Can be removed when we drop support for dagster_dg.library and dagster_dg.plugin
def _validate_plugin_entry_point(context: DgContext) -> None:
    if not context.pyproject_toml_path.exists():
        return

    import tomlkit

    toml = tomlkit.parse(context.pyproject_toml_path.read_text())
    for entry_point_group in OLD_DG_PLUGIN_ENTRY_POINT_GROUPS:
        if has_toml_node(toml, ("project", "entry-points", entry_point_group)):
            emit_warning(
                "deprecated_dagster_dg_library_entry_point",
                f"""
                Found deprecated `{entry_point_group}` entry point group in:
                    {context.pyproject_toml_path}

                Please update the group name to `dagster_dg_cli.registry_modules`. Package reinstallation is required
                because entry points are registered at install time. Reinstall your package to your
                environment using:

                    [uv]  $ uv pip install -e .
                    [pip] $ pip install -e .

                """,
                context.config.cli.suppress_warnings,
            )


DG_UPDATE_CHECK_INTERVAL = datetime.timedelta(hours=1)
DG_UPDATE_CHECK_ENABLED_ENV_VAR = "DAGSTER_DG_UPDATE_CHECK_ENABLED"


@whitelist_for_serdes
@record
class DgPyPiVersionInfo:
    timestamp: float
    raw_versions: list[str]

    @property
    def datetime(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.timestamp)

    @cached_property
    def versions(self) -> list[Version]:
        return sorted(Version(v) for v in self.raw_versions)



from pathlib import Path

from setuptools import find_packages, setup


def get_version() -> str:
    version: dict[str, str] = {}
    with open(Path(__file__).parent / "dagster_dg_core/version.py", encoding="utf8") as fp:
        exec(fp.read(), version)

    return version["__version__"]


ver = get_version()
# dont pin dev installs to avoid pip dep resolver issues
pin = "" if ver == "1!0+dev" else f"=={ver}"
setup(
    name="dagster-dg-core",
    version=get_version(),
    author="Dagster Labs",
    author_email="hello@dagsterlabs.com",
    license="Apache-2.0",
    description="",  # TODO - fill out description
    url=(
        "https://github.com/dagster-io/dagster/tree/master/python_modules/libraries/dagster-dg-core"
    ),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=["dagster_dg_core_tests*"]),
    install_requires=[
        "Jinja2",
        "tomlkit<0.13.3",  # bug in this version
        "click>=8,<8.2",
        "click-aliases",
        "typing_extensions>=4.4.0,<5",
        "gql[requests]",
        "markdown",
        "jsonschema",
        "PyYAML>=5.1",
        "rich",
        "watchdog",
        "yaspin",
        "setuptools",  # Needed to parse setup.cfg
        "packaging",
        "python-dotenv",
        "typer<0.17.0",
        "dagster-shared==1.11.8",
        "dagster-cloud-cli==1.11.8",
    ],
    include_package_data=True,
    zip_safe=False,
    entry_points={},
    extras_require={
        "test": [
            "click",
            "dagster==1.11.8",
            "freezegun",
            "psutil",
            "pydantic",
            "pytest",
            "dagster-graphql==1.11.8",
            f"create-dagster{pin}",
        ],
    },
)



import datetime
import time
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Callable, Optional

from dagster_shared.scaffold import DEFAULT_FILE_EXCLUDE_PATTERNS
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from dagster_dg_core.utils import clear_screen
from dagster_dg_core.utils.paths import hash_paths


class PathChangeHandler(FileSystemEventHandler):
    """Basic handler that clears the screen and re-executes the callback when any of the watched paths change
    in a way that would affect the hash of the paths. Passes the new hash to the callback.
    """

    def __init__(
        self,
        paths: Sequence[Path],
        includes: Optional[Sequence[str]],
        excludes: Sequence[str],
        callback: Callable[[str], Any],
    ):
        self._callback = callback
        self._paths = paths
        self._includes = includes
        self._excludes = excludes
        self._prev_hash = hash_paths(
            self._paths, self._includes, self._excludes, error_on_missing=False
        )
        self.clear_and_execute(self._prev_hash)

    def dispatch(self, _event: FileSystemEvent):  # pyright: ignore[reportIncompatibleMethodOverride]
        new_hash = hash_paths(self._paths, self._includes, self._excludes, error_on_missing=False)

        if new_hash != self._prev_hash:
            self.clear_and_execute(new_hash)
            self._prev_hash = new_hash

    def clear_and_execute(self, new_hash: str):
        clear_screen()
        self._callback(new_hash)
        current_time = datetime.datetime.now().strftime("%H:%M:%S")
        print(f"\nUpdated at {current_time}, watching for changes...")  # noqa: T201


# This is a global variable that is used to signal the watcher to exit in tests
SHOULD_WATCHER_EXIT = False


def watch_paths(
    paths: Sequence[Path],
    callback: Callable[[str], Any],
    includes: Optional[Sequence[str]] = None,
    excludes: Sequence[str] = DEFAULT_FILE_EXCLUDE_PATTERNS,
):
    """Watches the given paths for changes and calls the callback when they change.
    The callback should take a single argument, the new hash of the paths.

    Runs synchronously until the observer is stopped, or keyboard interrupt is received.

    Args:
        paths: The paths to watch.
        callback: The callback to call when path contents change.
        includes: A list of glob patterns to target, excluding files that don't match any of the patterns.
        excludes: A list of glob patterns to exclude, including files that match any of the patterns. Defaults to
            various Python-related files that are not relevant to the contents of the project.
    """
    observer = Observer()
    handler = PathChangeHandler(paths, includes, excludes, callback)
    for path in paths:
        observer.schedule(handler, str(path), recursive=True)
    observer.start()
    try:
        while observer.is_alive() and not SHOULD_WATCHER_EXIT:
            time.sleep(0.5)
    finally:
        observer.stop()
        observer.join()
